from pathlib import Path
from datetime import date, datetime
import re
import numpy as np

import pandas as pd
from pandas._testing import assert_frame_equal
import pytest

from py_ewr import scenario_handling, data_inputs

def test_match_MDBA_nodes():
    '''
    1. Ensure dataframe with flows and levels is split into two dataframes (one flow and one level dataframe)
    '''
    # Set up input data and pass to test function:
    model_metadata = data_inputs.get_MDBA_codes()
    data_df = {'Date': pd.date_range(start= datetime.strptime('2012-07-01', '%Y-%m-%d'), end = datetime.strptime('2016-06-30', '%Y-%m-%d')),
                'EUSTDS-1-8': [0]*1+[250]*350+[0]*9+[0]*5 + [0]*360+[0]*5 + [0]*2+[250]*345+[0]*1+[250]*17 + [0]*5+[250]*351+[250]*10,
                'EUSTUS-35-8': [0]*1+[250]*350+[0]*9+[0]*5 + [0]*360+[0]*5 + [0]*2+[250]*345+[0]*1+[250]*17 + [0]*5+[250]*351+[250]*10
                }
    df = pd.DataFrame(data = data_df)
    df = df.set_index('Date')
    
    df_F, df_L = scenario_handling.match_MDBA_nodes(df, model_metadata, 'py_ewr/parameter_metadata/parameter_sheet.csv')
    
    # Set up expected outputs and test:
    data_expected_df_L = {'Date': pd.date_range(start= datetime.strptime('2012-07-01', '%Y-%m-%d'), end = datetime.strptime('2016-06-30', '%Y-%m-%d')),
                            '414209': [0]*1+[250]*350+[0]*9+[0]*5 + [0]*360+[0]*5 + [0]*2+[250]*345+[0]*1+[250]*17 + [0]*5+[250]*351+[250]*10
                            }
    expected_df_L = pd.DataFrame(data_expected_df_L)
    expected_df_L = expected_df_L.set_index('Date') 
    data_expected_df_F = {'Date': pd.date_range(start= datetime.strptime('2012-07-01', '%Y-%m-%d'), end = datetime.strptime('2016-06-30', '%Y-%m-%d')),
                            '414203': [0]*1+[250]*350+[0]*9+[0]*5 + [0]*360+[0]*5 + [0]*2+[250]*345+[0]*1+[250]*17 + [0]*5+[250]*351+[250]*10
                            }
    expected_df_F = pd.DataFrame(data_expected_df_F)
    expected_df_F = expected_df_F.set_index('Date')
    
    assert_frame_equal(df_F, expected_df_F)
    assert_frame_equal(df_L, expected_df_L)

def test_match_NSW_nodes():
    '''
    1. Check NSW model nodes are mapped correctly to their gauges
    '''
    # Set up input data and pass to test function:
    model_metadata = data_inputs.get_NSW_codes()
    data_df = {'Date': pd.date_range(start= datetime.strptime('2012-07-01', '%Y-%m-%d'), end = datetime.strptime('2016-06-30', '%Y-%m-%d')),
                'Gauge: 410134 Billabong Creek at Darlot: Downstream Flow': [0]*1+[250]*350+[0]*9+[0]*5 + [0]*360+[0]*5 + [0]*2+[250]*345+[0]*1+[250]*17 + [0]*5+[250]*351+[250]*10,
                'Gauge: 410016 Billabong Creek at Jerilderie: Downstream Flow': [0]*1+[250]*350+[0]*9+[0]*5 + [0]*360+[0]*5 + [0]*2+[250]*345+[0]*1+[250]*17 + [0]*5+[250]*351+[250]*10
                }
    df = pd.DataFrame(data = data_df)
    df = df.set_index('Date')
    
    df_F, df_L = scenario_handling.match_NSW_nodes(df, model_metadata)
    
    # Set up expected outputs and test:
    data_expected_df_L = {'Date': pd.date_range(start= datetime.strptime('2012-07-01', '%Y-%m-%d'), end = datetime.strptime('2016-06-30', '%Y-%m-%d'))}
    expected_df_L = pd.DataFrame(data_expected_df_L)
    expected_df_L = expected_df_L.set_index('Date')
    
    data_expected_df_F = {'Date': pd.date_range(start= datetime.strptime('2012-07-01', '%Y-%m-%d'), end = datetime.strptime('2016-06-30', '%Y-%m-%d')),
                            '410134': [0]*1+[250]*350+[0]*9+[0]*5 + [0]*360+[0]*5 + [0]*2+[250]*345+[0]*1+[250]*17 + [0]*5+[250]*351+[250]*10,
                            '410016': [0]*1+[250]*350+[0]*9+[0]*5 + [0]*360+[0]*5 + [0]*2+[250]*345+[0]*1+[250]*17 + [0]*5+[250]*351+[250]*10
                            }
    expected_df_F = pd.DataFrame(data_expected_df_F)
    expected_df_F = expected_df_F.set_index('Date')
    
    assert_frame_equal(df_F, expected_df_F)
    assert_frame_equal(df_L, expected_df_L, check_column_type=False)

def test_extract_gauge_from_string():
    '''
    1. Check gauge string is pulled from various strings likely encountered during standard program run
    '''
    # Test 1
    input_string = '409025'
    return_gauge = scenario_handling.extract_gauge_from_string(input_string)
    expected_gauge = '409025'
    assert return_gauge == expected_gauge
    # Test 2 - TODO - is this test necessary?
    # input_string = ' 409025 '
    # return_gauge = scenario_handling.extract_gauge_from_string(input_string)
    # expected_gauge = '409025'
    # assert return_gauge == expected_gauge
    # Test 3
    # input_string = '409025---999'
    # return_gauge = scenario_handling.extract_gauge_from_string(input_string)
    # expected_gauge = '409025'
    # assert return_gauge == expected_gauge

    

## changed to 100 year
def test_cleaner_standard_timeseries():
    '''
    1. Check date formatting and correct allocationo of gauge data to either flow/level dataframes
    2. TODO: Add test to assess when there are missing dates
    3. TODO: Add test for both date types
    '''
    # CHECK 1
    # Set up input data and pass to test function:
    date_start = '1900-07-01'
    date_end = '2000-06-30'
    date_range = pd.date_range(start= datetime.strptime(date_start, '%Y-%m-%d'), end = datetime.strptime(date_end, '%Y-%m-%d'), freq='D')
 
    data_for_input_df = {'Date': date_range, '409025_flow': [50]*len(date_range)}
    input_df = pd.DataFrame(data_for_input_df)
    input_df = input_df.set_index('Date')
    df_f, df_l = scenario_handling.cleaner_standard_timeseries(input_df)

    # Set up expected data and test:
    expected_df_flow = input_df.copy(deep=True)
    expected_df_flow.columns = ['409025']
    expected_df_level = pd.DataFrame(index = expected_df_flow.index)

    expected_df_flow.index = date_range
    expected_df_flow.index.name = 'Date'

    assert_frame_equal(expected_df_level, df_l)
    assert_frame_equal(expected_df_flow, df_f)

    # CHECK 2

    date_start = '1900-07-01'
    date_end = '2000-06-30'
    date_range = pd.date_range(start= datetime.strptime(date_start, '%Y-%m-%d'), end = datetime.strptime(date_end, '%Y-%m-%d'), freq='D')

    data_for_input_df = {'Date': date_range, '409025_flow': [50]*len(date_range)}
    input_df = pd.DataFrame(data_for_input_df)
    input_df = input_df.set_index('Date')
    cut_df = input_df.drop([datetime(1900, 7, 5),datetime(1900,7,8)], axis = 0)

    df_f, df_l = scenario_handling.cleaner_standard_timeseries(cut_df)
    
    expected_df_flow = cut_df.copy(deep = True)
    expected_df_flow.loc[datetime(1900, 7, 5), '409025_flow'] = np.nan
    expected_df_flow.loc[datetime(1900, 7, 8), '409025_flow'] = np.nan
    expected_df_flow.sort_index(inplace=True)
    expected_df_flow.columns = ['409025']
    expected_df_level = pd.DataFrame(index = expected_df_flow.index)

    expected_df_flow.index = date_range
    expected_df_flow.index.name = 'Date'

    assert_frame_equal(expected_df_level, df_l)
    assert_frame_equal(expected_df_flow, df_f)

    
def test_cleaner_NSW():
    '''
    1. Test date formatting is applied correctly
    '''
    # Set up input data and pass to test function:
    model_metadata = data_inputs.get_NSW_codes()
    dates = pd.date_range(start= datetime.strptime('2012-07-01', '%Y-%m-%d'), end = datetime.strptime('2016-06-30', '%Y-%m-%d'))
    date_strings = dates.format(formatter=lambda x: x.strftime('%Y-%m-%d'))
    data_df = {'Date': date_strings,
                'Gauge: YCB_410134_BillabongCreek@Darlot: Downstream Flow': [0]*1+[250]*350+[0]*9+[0]*5 + [0]*360+[0]*5 + [0]*2+[250]*345+[0]*1+[250]*17 + [0]*5+[250]*351+[250]*10,
                'Gauge: YCB_410016 Billabong Creek @ Jerilderie: Downstream Flow': [0]*1+[250]*350+[0]*9+[0]*5 + [0]*360+[0]*5 + [0]*2+[250]*345+[0]*1+[250]*17 + [0]*5+[250]*351+[250]*10
                }
    df = pd.DataFrame(data = data_df)
    df_clean = scenario_handling.cleaner_NSW(df)
    # Set up expected output data and test:
    expected_dates = pd.date_range(start= datetime.strptime('2012-07-01', '%Y-%m-%d'), end = datetime.strptime('2016-06-30', '%Y-%m-%d'))
    data_expected_df_F = {'Date': expected_dates,
                            'Gauge: YCB_410134_BillabongCreek@Darlot: Downstream Flow': [0]*1+[250]*350+[0]*9+[0]*5 + [0]*360+[0]*5 + [0]*2+[250]*345+[0]*1+[250]*17 + [0]*5+[250]*351+[250]*10,
                            'Gauge: YCB_410016 Billabong Creek @ Jerilderie: Downstream Flow': [0]*1+[250]*350+[0]*9+[0]*5 + [0]*360+[0]*5 + [0]*2+[250]*345+[0]*1+[250]*17 + [0]*5+[250]*351+[250]*10
                            }
    expected_df_F = pd.DataFrame(data_expected_df_F)
    expected_df_F['Date'] = expected_df_F['Date'].apply(lambda x: x.to_period(freq='D'))
    expected_df_F = expected_df_F.set_index('Date')
    assert_frame_equal(df_clean, expected_df_F)
    
    
def test_cleaner_MDBA():
    '''
    1. Test date formatting is applied correctly
    '''
    # Set up input data and pass to test function:
    data_df = {'Date': pd.date_range(start= datetime.strptime('2012-07-01', '%Y-%m-%d'), end = datetime.strptime('2016-06-30', '%Y-%m-%d')),
                'EUSTDS-1-8': [0]*1+[250]*350+[0]*9+[0]*5 + [0]*360+[0]*5 + [0]*2+[250]*345+[0]*1+[250]*17 + [0]*5+[250]*351+[250]*10,
                'EUSTUS-35-8': [0]*1+[250]*350+[0]*9+[0]*5 + [0]*360+[0]*5 + [0]*2+[250]*345+[0]*1+[250]*17 + [0]*5+[250]*351+[250]*10
                }
    df = pd.DataFrame(data = data_df)
    df['Dy'], df['Mn'], df['Year'] = df['Date'].dt.day, df['Date'].dt.month, df['Date'].dt.year
    df = df.drop(['Date'], axis = 1)
    
    df_clean = scenario_handling.cleaner_MDBA(df)
    # Set up expected output data and test:
    data_expected_df = {'Date': pd.date_range(start= datetime.strptime('2012-07-01', '%Y-%m-%d'), end = datetime.strptime('2016-06-30', '%Y-%m-%d')).to_period(),
                            'EUSTDS-1-8': [0]*1+[250]*350+[0]*9+[0]*5 + [0]*360+[0]*5 + [0]*2+[250]*345+[0]*1+[250]*17 + [0]*5+[250]*351+[250]*10,
                            'EUSTUS-35-8': [0]*1+[250]*350+[0]*9+[0]*5 + [0]*360+[0]*5 + [0]*2+[250]*345+[0]*1+[250]*17 + [0]*5+[250]*351+[250]*10
                            }
    expected_df = pd.DataFrame(data_expected_df)
    expected_df = expected_df.set_index('Date')
    
    assert_frame_equal(df_clean, expected_df)
    
    
def test_build_NSW_columns():
    '''
    1. Check data columns are renamed based on header data
    '''
    # Load in test input data and pass to test function:
    input_file = 'unit_testing_files/NSW_source_res_test_file.csv'
    data, header = scenario_handling.unpack_model_file(input_file, 'Date', 'Field')
    data = scenario_handling.build_NSW_columns(data, header)
    # Load in test expected output data and test:
    expected_result = pd.read_csv('unit_testing_files/NSW_source_res_test_file_flow_result.csv')
    assert_frame_equal(data, expected_result)
    
    
def test_build_MDBA_columns():
    '''
    1. Ensure data column names are correctly taken from the header data
    '''
    # Load input data and send to test function:
    input_file = 'unit_testing_files/MDBA_bigmod_test_file_buggy_ID.csv'
    data, header = scenario_handling.unpack_model_file(input_file, 'Dy', 'Field')
    data = scenario_handling.build_MDBA_columns(data, header)
    data = data.astype({'Dy': 'int32', 'Mn': 'int32', 'Year': 'int32'})
    # Load expected output data, format, and test:
    expected_result = pd.read_csv('unit_testing_files/MDBA_bigmod_test_file_flow_result.csv')
    expected_result = expected_result.astype({'Dy': 'int32', 'Mn': 'int32', 'Year': 'int32'})
    assert_frame_equal(data, expected_result)        


def test_unpack_model_file():
    '''
    1. Test MDBA style file ingestion
    2. Test NSW style file ingestion
    '''
    
    # Test 1
    # Load test data and expected output data
    file_to_pass = 'unit_testing_files/MDBA_bigmod_test_file.csv'
    expected_header = pd.read_csv('unit_testing_files/MDBA_bigmod_test_file_header_result.csv', dtype={'Site':'str', 'Measurand': 'str', 'Quality': 'str'})
    expected_flow = pd.read_csv('unit_testing_files/MDBA_bigmod_test_file_flow_result.csv', dtype={'Dy':'int', 'Mn': 'int', 'Year': 'int'})
    # Pass to test function and test
    flow, header = scenario_handling.unpack_model_file(file_to_pass, 'Dy', 'Field')
    assert_frame_equal(flow, expected_flow)
    assert_frame_equal(header, expected_header)
    #------------------------------------------
    # Test 2
    # Load test data and expected output data:
    file_to_pass = 'unit_testing_files/NSW_source_res_test_file.csv'
    expected_header = pd.read_csv('unit_testing_files/NSW_source_res_test_file_header_result.csv')
    expected_flow = pd.read_csv('unit_testing_files/NSW_source_res_test_file_flow_result.csv')
    expected_flow.columns = ['Date', '1>Data Sources>Data Sources@Climate Data@FAO56_res_csv@049050_SILO_FAO56.csv', '2>Data Sources>Data Sources@Climate Data@Mwet_res_csv@049050_SILO_Mwet.csv']
    # Pass to test function and test
    flow, header = scenario_handling.unpack_model_file(file_to_pass, 'Date', 'Field')  
    assert_frame_equal(flow, expected_flow)
    assert_frame_equal(header, expected_header)

    
# def test_gauge_only_column():
#     '''
#     1. Check gauge strings are pulled from column names and saved in place of the original string 
#     '''
#     file_to_pass ='unit_testing_files/NSW_10000yr_test_file.csv'
#     flow = scenario_handling.gauge_only_column(file_to_pass)
    
#     expected_flow = pd.read_csv('unit_testing_files/NSW_10000yr_test_file.csv', index_col = 'Date')
#     expected_flow.columns = ['418013']
    
#     assert_frame_equal(flow, expected_flow)

def test_scenario_handler_class(scenario_handler_expected_detail, scenario_handler_instance):
   
    detailed = scenario_handler_instance.pu_ewr_statistics
    
    detailed['Low_flow_EWRs_Bidgee_410007']['410007']['Upper Yanco Creek'].index = detailed['Low_flow_EWRs_Bidgee_410007']['410007']['Upper Yanco Creek'].index.astype('int64')

    assert_frame_equal(detailed['Low_flow_EWRs_Bidgee_410007']['410007']['Upper Yanco Creek'], scenario_handler_expected_detail)


def test_get_all_events(scenario_handler_instance):

    all_events = scenario_handler_instance.get_all_events()
    assert type(all_events) == pd.DataFrame
    assert all_events.shape == (26, 10)
    assert all_events.columns.to_list() == ['scenario', 'gauge', 'pu', 'ewr', 'waterYear', 'startDate', 'endDate',
                                     'eventDuration', 'eventLength', 'Multigauge']
        
def test_get_all_interEvents(scenario_handler_instance):

    all_interEvents = scenario_handler_instance.get_all_interEvents()
    assert type(all_interEvents) == pd.DataFrame
    assert all_interEvents.shape == (26, 7)
    assert all_interEvents.columns.to_list() == ['scenario', 'gauge', 'pu', 'ewr', 'startDate', 'endDate', 'interEventLength']

def test_get_all_successful_events(scenario_handler_instance):

    all_successful_events = scenario_handler_instance.get_all_successful_events()
    assert type(all_successful_events) == pd.DataFrame
    assert all_successful_events.shape == (24, 10)
    assert all_successful_events.columns.to_list() == ['scenario', 'gauge', 'pu', 'ewr', 'waterYear', 'startDate', 'endDate',
                                                       'eventDuration', 'eventLength', 'Multigauge']

def test_get_all_successful_interEvents(scenario_handler_instance):

    all_successful_interEvents = scenario_handler_instance.get_all_successful_interEvents()
    assert type(all_successful_interEvents) == pd.DataFrame
    assert all_successful_interEvents.shape == (24, 7)
    assert all_successful_interEvents.columns.to_list() == ['scenario', 'gauge', 'pu', 'ewr', 'startDate', 'endDate', 'interEventLength']

def test_get_yearly_ewr_results(scenario_handler_instance):

    yearly_results = scenario_handler_instance.get_yearly_ewr_results()

    assert type(yearly_results) == pd.DataFrame
    assert yearly_results.shape == (126, 21)
    assert yearly_results.columns.to_list() == ['Year', 'eventYears', 'numAchieved', 'numEvents', 'numEventsAll',
        'eventLength', 'eventLengthAchieved',
       'totalEventDays', 'totalEventDaysAchieved','maxEventDays', 'maxRollingEvents', 'maxRollingAchievement', 'missingDays',
       'totalPossibleDays', 'ewrCode', 'scenario', 'gauge', 'pu', 'Multigauge', 
       'rollingMaxInterEvent', 'rollingMaxInterEventAchieved'] 

def test_get_ewr_results(scenario_handler_instance):

    ewr_results = scenario_handler_instance.get_ewr_results()
    assert type(ewr_results) == pd.DataFrame
    assert ewr_results.shape == (21, 19)
    assert ewr_results.columns.to_list() == ['Scenario', 'Gauge', 'PlanningUnit', 'EwrCode', 'Multigauge','EventYears',
       'Frequency', 'TargetFrequency', 'AchievementCount',
       'AchievementPerYear', 'EventCount', 'EventCountAll', 'EventsPerYear', 'EventsPerYearAll',
       'AverageEventLength', 'ThresholdDays', 
       'MaxInterEventYears', 'NoDataDays', 'TotalDays']
    
def test_any_cllmm_to_process(gauge_results):
    result = scenario_handling.any_cllmm_to_process(gauge_results)
    assert result == True

    


from typing import Dict, List
import csv
import os
import urllib
import re
from datetime import datetime, date
import logging

import pandas as pd
from tqdm import tqdm
 
log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


from . import data_inputs, evaluate_EWRs, summarise_results
#----------------------------------- Scenario testing handling functions--------------------------#

def unpack_model_file(csv_file: str, main_key: str, header_key: str) -> tuple:
    '''Ingesting scenario file locations of model files with all formats (excluding standard timeseries format), seperates the flow data and header data
    returns a dictionary of flow dataframes with their associated header data
    
    Args:
        csv_file (str): location of model file
        main_key (str): unique identifier for the start of the flow data (dependent on model format type being uploaded)
        header_key (str): unique identifier for the start of the header data (dependent on model format type being uploaded)
    
    Results:
        tuple[pd.DataFrame, pd.DataFrame]: flow dataframe; header dataframe
    
    '''
    if csv_file[-3:] != 'csv':
        raise ValueError('''Incorrect file type selected, bigmod format requires a csv file.
                         Rerun the program and try again.''')
    
    #--------functions for pulling main data-------#
    
    def mainData_url(url, line,**kwargs):
        '''Get daily data (excluding the header data); remote file upload'''
        response = urllib.request.urlopen(url)
        lines = [l.decode('utf-8') for l in response.readlines()]
        cr = csv.reader(lines)
        pos = 0
        for row in cr:
            if row[0].startswith(line):
                headerVal = pos
                break
            pos = pos + 1
        if main_key == 'Dy':
            df = pd.read_csv(url, header=headerVal, dtype={'Dy':'int', 'Mn': 'int', 'Year': 'int'}, skiprows=range(headerVal+1, headerVal+2))
        elif main_key == 'Date':
            df = pd.read_csv(url, header=headerVal, skiprows=range(headerVal+1, headerVal+2))
        df = df.loc[:, ~df.columns.str.contains('^Unnamed')]
        
        return df, headerVal
        
    def mainData(file, line,**kwargs):
        '''Get daily data (excluding the header data); local file upload'''
        if os.stat(file).st_size == 0:
            raise ValueError("File is empty")
        with open(file) as csv_file:
            csv_reader = csv.reader(csv_file) #, delimiter=','
            line_count = 0
            for row in csv_reader:
                if row[0].startswith(line):
                    headerVal = line_count
                    break
                line_count = line_count + 1
        if main_key == 'Dy':
            df = pd.read_csv(file, header=headerVal, dtype={'Dy':'int', 'Mn': 'int', 'Year': 'int'}, skiprows=range(headerVal+1, headerVal+2))
        elif main_key == 'Date':
            df = pd.read_csv(file, header=headerVal, skiprows=range(headerVal+1, headerVal+2))
        df = df.loc[:, ~df.columns.str.contains('^Unnamed')]

        return df, headerVal
        
    #--------Functions for pulling header data---------#
    
    def headerData_url(url, line, endLine, **kwargs):
        '''Get header data for a remote file upload'''
        response = urllib.request.urlopen(url)
        lines = [l.decode('utf-8') for l in response.readlines()]
        cr = csv.reader(lines)
        pos = 0
        for row in cr:
            if row[0].startswith(line):
                headerVal = pos
                break
            pos = pos + 1
        junkRows = headerVal # Junk rows because rows prior to this value will be discarded
        df = pd.read_csv(url, header=headerVal, nrows = (endLine-junkRows-1), dtype={'Site':'str', 'Measurand': 'str', 'Quality': 'str'})
        df = df.loc[:, ~df.columns.str.contains('^Unnamed')]
        
        return df
        
    def headerData(file, line, endLine, **kwargs):
        '''Get header data for local file upload'''
        if os.stat(file).st_size == 0:
            raise ValueError("File is empty")
            
        with open(file) as csv_file:
            csv_reader = csv.reader(csv_file)
            line_count = 0
            for row in csv_reader:
                if row[0].startswith(line):
                    headerVal = line_count
                    # Then get column length:
                    num_cols = num_cols = list(range(0,len(row),1))
                    break
                    
                line_count = line_count + 1
            junkRows = headerVal # Junk rows because rows prior to this value will be discarded
            
        df = pd.read_csv(file, header=headerVal, usecols = num_cols, 
                         nrows = (endLine-junkRows-1), dtype={'Site':'str', 'Measurand': 'str', 'Quality': 'str'},
                        encoding = 'utf-8', sep =',')
        df = df.loc[:, ~df.columns.str.contains('^Unnamed')]
            
        return df
    
    if 'http' in csv_file:
        mainData_df, endLine = mainData_url(csv_file, main_key, sep=",")
        headerData_df = headerData_url(csv_file, header_key, endLine, sep=",")
    else:
        mainData_df, endLine = mainData(csv_file, main_key, sep=",")
        headerData_df = headerData(csv_file, header_key, endLine, sep=",")
    
    return mainData_df, headerData_df

def build_MDBA_columns(input_data: pd.DataFrame, input_header: pd.DataFrame) -> pd.DataFrame:
    '''Takes in the header data file, trims it, and then renames the column headings with the full reference code
    returns a the dataframe with updated column headers (MDBA model formats).
    
    Args:
        input_data (pd.DataFrame): flow/water level dataframe
        input_header (pd.DataFrame): header dataframe

    Results:
        pd.DataFrame: the flow dataframe with updated header data
    
    '''
    # Clean dataframe
    numRows = int(input_header['Field'].iloc[1]) 
    df = input_header.drop([0,1])     
    df = df.astype(str)
    # Remove rogue quotes, spaces, and apostrophes
    df['Site'] = df['Site'].map(lambda x: x.replace("'", ""))
    df['Site'] = df['Site'].map(lambda x: x.replace('"', ''))
    df['Site'] = df['Site'].map(lambda x: x.replace(" ", ""))

    df['Measurand'] = df['Measurand'].map(lambda x: x.replace("'", ""))
    df['Measurand'] = df['Measurand'].map(lambda x: x.replace('"', ''))
    df['Measurand'] = df['Measurand'].map(lambda x: x.replace(" ", ""))

    df['Quality'] = df['Quality'].map(lambda x: x.replace("'", ""))
    df['Quality'] = df['Quality'].map(lambda x: x.replace('"', ''))
    df['Quality'] = df['Quality'].map(lambda x: x.replace(" ", ""))

    # Construct refs and save to list:
    listOfCols = []
    for i in range(0, numRows):
        colName = str(df['Site'].iloc[i] + '-' + df['Measurand'].iloc[i] + '-' + df['Quality'].iloc[i])
        listOfCols.append(colName)
    dateList = ['Dy', 'Mn', 'Year']
    listOfCols = dateList + listOfCols

    input_data.columns = listOfCols
    
    return input_data

def build_NSW_columns(input_data: pd.DataFrame, input_header: pd.DataFrame) -> pd.DataFrame:
    '''Takes in the header data file, trims it, and then renames the column headings with the full reference code
    returns a the dataframe with updated column headers (NSW res.csv model format).
    
    Args:
        input_data (pd.DataFrame): flow/water level dataframe
        input_header (pd.DataFrame): header dataframe

    Results:
        pd.DataFrame: the flow/water level dataframe with updated header data

    '''
    # Extract unique column ID's from the header:
    
    new_cols = input_header['Name'].to_list()
    new_cols = new_cols[2:]
    new_cols.insert(0, 'Date')
    input_data.columns = new_cols
    
    return input_data

def cleaner_MDBA(input_df: pd.DataFrame) -> pd.DataFrame:
    '''Ingests dataframe, removes junk columns, fixes date,
    returns formatted dataframe
    
    Args:
        input_df (pd.DataFrame): flow/water level dataframe
    Results:
        pd.DataFrame: Cleaned flow/water level dataframe
    
    '''
    
    cleaned_df = input_df.rename(columns={'Mn': 'Month', 'Dy': 'Day'})
    cleaned_df['Date'] = pd.to_datetime(cleaned_df[['Year', 'Month', 'Day']], format = '%Y-%m-%d')
    cleaned_df['Date'] = cleaned_df['Date'].apply(lambda x: x.to_period(freq='D'))
    cleaned_df = cleaned_df.drop(['Day', 'Month', 'Year'], axis = 1)
    cleaned_df = cleaned_df.set_index('Date')
    
    return cleaned_df

def cleaner_NSW(input_df: pd.DataFrame) -> pd.DataFrame:
    '''Convert dates to datetime format, save this to the dataframe index
    
    Args:
        input_df (pd.DataFrame): flow/water level dataframe
    Results:
        pd.DataFrame: Cleaned flow/water level dataframe

    '''
    
    cleaned_df = input_df.copy(deep=True)
    
    try:
        cleaned_df['Date'] = pd.to_datetime(cleaned_df['Date'], format = '%d/%m/%Y')
        cleaned_df['Date'] = cleaned_df['Date'].apply(lambda x: x.to_period(freq='D'))
    except ValueError:
        log.info('''Attempted and failed to read in dates in format: dd/mm/yyyy, 
        attempting to look for dates in format: yyyy-mm-dd''')
        try:
            cleaned_df['Date'] = pd.to_datetime(cleaned_df['Date'], format = '%Y-%m-%d')
            cleaned_df['Date'] = cleaned_df['Date'].apply(lambda x: x.to_period(freq='D'))
        except ValueError:
            raise ValueError('New date format detected. Cannot read in data')
        log.info('successfully read in data with yyyy-mm-dd formatting')
    cleaned_df = cleaned_df.set_index('Date')
    
    return cleaned_df

def cleaner_standard_timeseries(input_df: pd.DataFrame, ewr_table_path: str = None) -> pd.DataFrame:
    '''Ingests dataframe, removes junk columns, fixes date, allocates gauges to either flow/level
    
    Args:
        input_df (pd.DataFrame): flow/water level dataframe

    Results:
        tuple[pd.DataFrame, pd.DataFrame]: Cleaned flow dataframe; cleaned water level dataframe

    '''

    cleaned_df = input_df.copy(deep=True)
    try:
        cleaned_df.index = pd.to_datetime(cleaned_df.index, format = '%d/%m/%Y')
    except ValueError:
        log.info('''Attempted and failed to read in dates in format: dd/mm/yyyy, attempting
        to look for dates in format: yyyy-mm-dd''')
        try:
            cleaned_df.index = pd.to_datetime(cleaned_df.index, format = '%Y-%m-%d')
        except ValueError:
            raise ValueError('''New date format detected. Cannot read in data''')
        log.info('''Successfully read in data with yyyy-mm-dd formatting''')

    # If there are missing dates, add in the dates and fill with NaN values
    dates = pd.date_range(start = cleaned_df.index[0], end=cleaned_df.index[-1])
    cleaned_df = cleaned_df.reindex(dates)

    # TODO: Optional: add in gap filling code if user selects preference for gap filling

    df_flow = pd.DataFrame(index = cleaned_df.index)
    df_level = pd.DataFrame(index = cleaned_df.index)
    df_flow.index.name = 'Date'
    df_level.index.name = 'Date'

    for gauge in cleaned_df.columns:
        gauge_only = extract_gauge_from_string(gauge)
        if 'flow' in gauge:
            df_flow[gauge_only] = cleaned_df[gauge].copy(deep=True)
        if 'level' in gauge:
            df_level[gauge_only] = cleaned_df[gauge].copy(deep=True)
        if not gauge_only:
            log.info('Could not identify gauge in column name:', gauge, ', skipping analysis of data in this column.')
    return df_flow, df_level


def cleaner_ten_thousand_year(input_df: pd.DataFrame, ewr_table_path: str = None) -> pd.DataFrame:
    '''Ingests dataframe, removes junk columns, fixes date, allocates gauges to either flow/level
    
    Args:
        input_df (pd.DataFrame): flow/water level dataframe

    Results:
        tuple[pd.DataFrame, pd.DataFrame]: Cleaned flow dataframe; cleaned water level dataframe

    '''
    
    cleaned_df = input_df.copy(deep=True)
    
    try:
        date_start = datetime.strptime(cleaned_df.index[0], '%d/%m/%Y')
        date_end = datetime.strptime(cleaned_df.index[-1], '%d/%m/%Y')
    except ValueError:    
        log.info('Attempted and failed to read in dates in format: dd/mm/yyyy, attempting to look for dates in format: yyyy-mm-dd')
        try:
            date_start = datetime.strptime(cleaned_df.index[0], '%Y-%m-%d')
            date_end = datetime.strptime(cleaned_df.index[-1], '%Y-%m-%d')
        except ValueError:
            raise ValueError('New date format detected. Cannot read in data')
        log.info('successfully read in data with yyyy-mm-dd formatting')
    date_range = pd.period_range(date_start, date_end, freq = 'D')
    cleaned_df['Date'] = date_range
    cleaned_df = cleaned_df.set_index('Date')

    df_flow = pd.DataFrame(index = cleaned_df.index)
    df_level = pd.DataFrame(index = cleaned_df.index)

    for gauge in cleaned_df.columns:
        gauge_only = extract_gauge_from_string(gauge)
        if 'flow' in gauge:
            df_flow[gauge_only] = cleaned_df[gauge].copy(deep=True)
        if 'level' in gauge:
            df_level[gauge_only] = cleaned_df[gauge].copy(deep=True)
        if not gauge_only:
            log.info('Could not identify gauge in column name:', gauge, ', skipping analysis of data in this column.')
    return df_flow, df_level

def extract_gauge_from_string(input_string: str) -> str:
    '''Takes in a strings, pulls out the gauge number from this string
    
    Args:
        input_string (str): string which may contain a gauge number

    Returns:
        str: Gauge number as a string if found, None if not found
    '''
    gauge = input_string.split('_')[0]
    return gauge

def match_MDBA_nodes(input_df: pd.DataFrame, model_metadata: pd.DataFrame, ewr_table_path: str) -> tuple:
    '''Checks if the source file columns have EWRs available, returns a flow and level dataframe with only 
    the columns with EWRs available. Renames columns to gauges
    
    Args:
        input_df (pd.DataFrame): flow/water level dataframe
        model_metadata (pd.DataFrame): dataframe linking model nodes to gauges

    Returns:
        tuple[pd.DataFrame, pd.DataFrame]: flow dataframe, water level dataframe

    '''

    flow_gauges = data_inputs.get_gauges('flow gauges', ewr_table_path=ewr_table_path)
    level_gauges = data_inputs.get_gauges('level gauges', ewr_table_path=ewr_table_path)
    measurands = ['1', '35']
    df_flow = pd.DataFrame(index = input_df.index)
    df_level = pd.DataFrame(index = input_df.index)
    for col in input_df.columns:
        col_clean = col.replace(' ', '')
        site = col_clean.split('-')[0]
        measure = col_clean.split('-')[1]
        if ((measure in measurands) and (model_metadata['SITEID'] == site).any()):
            subset = model_metadata.query("SITEID==@site")
            gauge = subset["AWRC"].iloc[0]
            if gauge in flow_gauges and measure == '1':
                df_flow[gauge] = input_df[col]
            if gauge in level_gauges and measure == '35':
                df_level[gauge] = input_df[col]
    if df_flow.empty:
        raise ValueError('No relevant gauges and or measurands found in dataset, the EWR tool cannot evaluate this model output file')      
    return df_flow, df_level

def match_NSW_nodes(input_df: pd.DataFrame, model_metadata: pd.DataFrame) -> tuple:
    '''Checks if the source file columns have EWRs available, returns a flow and level dataframe with only 
    the columns with EWRs available. Renames columns to gauges
    
    Args:
        input_df (pd.DataFrame): flow/water level dataframe
        model_metadata (pd.DataFrame): dataframe linking model nodes to gauges

    Returns:
        tuple[pd.DataFrame, pd.DataFrame]: flow dataframe, water level dataframe
    
    '''
    flow_gauges = data_inputs.get_gauges('flow gauges')
    level_gauges = data_inputs.get_gauges('level gauges')
    df_flow = pd.DataFrame(index = input_df.index)
    df_level = pd.DataFrame(index = input_df.index)
    for col in input_df.columns:
        if not (model_metadata[model_metadata['SITEID'].str.contains(col, na=False, regex=False)]).empty:
            subset = model_metadata.query("SITEID==@col")
            gauge = subset["AWRC"].iloc[0]
            if gauge in flow_gauges:
                df_flow[gauge] = input_df[col]
            if gauge in level_gauges:
                df_level[gauge] = input_df[col]

    return df_flow, df_level

def any_cllmm_to_process(gauge_results: dict)->bool:
    cllmm_gauges = data_inputs.get_cllmm_gauges()
    processed_gauges = data_inputs.get_scenario_gauges(gauge_results)
    return any(gauge in processed_gauges for gauge in cllmm_gauges)

class ScenarioHandler:
    
    def __init__(self, scenario_file: str, model_format:str, parameter_sheet:str = None,
                calc_config_path:str = None):
        self.scenario_file = scenario_file
        self.model_format = model_format
        self.yearly_events = None
        self.pu_ewr_statistics = None
        self.summary_results = None
        self.parameter_sheet = parameter_sheet
        self.calc_config_path = calc_config_path
        self.flow_data = None
        self.level_data = None

    def _get_file_names(self, loaded_files):

        file_locations = {}
        # for file in loaded_files:
        if '/' in loaded_files:
            full_name = loaded_files.split('/')
        elif ('\\' in loaded_files):
            full_name = loaded_files.split('\\')
        else:
            full_name = loaded_files
        name_exclude_extension = full_name[-1].split('.csv')[0]
        file_locations[str(name_exclude_extension)] = loaded_files
            
        return file_locations
        
    def process_scenarios(self):

        scenarios = self._get_file_names(self.scenario_file)

        # Analyse all scenarios for EWRs
        detailed_results = {}
        detailed_events = {}
        for scenario in tqdm(scenarios, position = 0, leave = True, 
                            bar_format='{l_bar}{bar:10}{r_bar}{bar:-10b}',
                            desc= 'Evaluating scenarios'):
            if self.model_format == 'Bigmod - MDBA':
                
                data, header = unpack_model_file(scenarios[scenario], 'Dy', 'Field')
                data = build_MDBA_columns(data, header)
                df_clean = cleaner_MDBA(data)
                df_F, df_L = match_MDBA_nodes(df_clean, data_inputs.get_MDBA_codes(), self.parameter_sheet)
                               
            elif self.model_format == 'Standard time-series':
                df = pd.read_csv(scenarios[scenario], index_col = 'Date')
                df_F, df_L = cleaner_standard_timeseries(df, self.parameter_sheet)

            elif self.model_format == 'Source - NSW (res.csv)':
                data, header = unpack_model_file(scenarios[scenario], 'Date', 'Field')
                data = build_NSW_columns(data, header)
                df_clean = cleaner_NSW(data)
                df_F, df_L = match_NSW_nodes(df_clean, data_inputs.get_NSW_codes())

            elif self.model_format == 'ten thousand year':
                df = pd.read_csv(scenarios[scenario], index_col = 'Date')
                df_F, df_L = cleaner_ten_thousand_year(df, self.parameter_sheet)
            
            gauge_results = {}
            gauge_events = {}

            all_locations = set(df_F.columns.to_list() + df_L.columns.to_list())
            EWR_table, bad_EWRs = data_inputs.get_EWR_table(self.parameter_sheet)
            calc_config = data_inputs.get_ewr_calc_config(self.calc_config_path)
            for gauge in all_locations:
                gauge_results[gauge], gauge_events[gauge] = evaluate_EWRs.calc_sorter(df_F, df_L, gauge,
                                                                                        EWR_table, calc_config) 
            detailed_results[scenario] = gauge_results
            detailed_events[scenario] = gauge_events
            self.pu_ewr_statistics = detailed_results
            self.yearly_events = detailed_events
            
            self.flow_data = df_F
            self.level_data = df_L

    def get_all_events(self)-> pd.DataFrame:

        if not self.yearly_events:
            self.process_scenarios()
        
        events_to_process = summarise_results.get_events_to_process(self.yearly_events)
        all_events = summarise_results.process_all_events_results(events_to_process)

        all_events = summarise_results.join_ewr_parameters(cols_to_add=['Multigauge'],
                        left_table=all_events,
                        left_on=['gauge','pu','ewr'],
                        selected_columns= ['scenario', 'gauge', 'pu', 'ewr', 'waterYear', 'startDate', 'endDate',
                                'eventDuration', 'eventLength', 
                                'Multigauge'],
                        parameter_sheet_path=self.parameter_sheet)

        all_events = summarise_results.filter_duplicate_start_dates(all_events)

        return all_events

    def get_all_interEvents(self)-> pd.DataFrame:
        
        if not self.yearly_events:
            self.process_scenarios()
        
        events_to_process = summarise_results.get_events_to_process(self.yearly_events)
        all_events_temp = summarise_results.process_all_events_results(events_to_process)

        all_events_temp = summarise_results.join_ewr_parameters(cols_to_add=['Multigauge'],
                        left_table=all_events_temp,
                        left_on=['gauge','pu','ewr'],
                        selected_columns= ['scenario', 'gauge', 'pu', 'ewr', 'waterYear', 'startDate', 'endDate',
                                'eventDuration', 'eventLength', 
                                'Multigauge'],
                        parameter_sheet_path=self.parameter_sheet)

        all_events_temp = summarise_results.filter_duplicate_start_dates(all_events_temp)

        # Get start and end date of the timeseries.
        date0 = self.flow_data.index[0]
        date1 = self.flow_data.index[-1]
        start_date = date(date0.year, date0.month, date0.day)
        end_date = date(date1.year, date1.month, date1.day)
        
        all_interEvents = summarise_results.events_to_interevents(start_date, end_date, all_events_temp)

        return all_interEvents        

    def get_all_successful_events(self)-> pd.DataFrame:

        if not self.yearly_events:
            self.process_scenarios()
        
        events_to_process = summarise_results.get_events_to_process(self.yearly_events)
        all_events_temp1 = summarise_results.process_all_events_results(events_to_process)

        all_events_temp1 = summarise_results.join_ewr_parameters(cols_to_add=['Multigauge'],
                        left_table=all_events_temp1,
                        left_on=['gauge','pu','ewr'],
                        selected_columns= ['scenario', 'gauge', 'pu', 'ewr', 'waterYear', 'startDate', 'endDate',
                                'eventDuration', 'eventLength', 
                                'Multigauge'],
                        parameter_sheet_path=self.parameter_sheet)

        all_events_temp1 = summarise_results.filter_duplicate_start_dates(all_events_temp1)

        all_successfulEvents = summarise_results.filter_successful_events(all_events_temp1, self.parameter_sheet) 

        return all_successfulEvents

    def get_all_successful_interEvents(self)-> pd.DataFrame:

        if not self.yearly_events:
            self.process_scenarios()
        
        events_to_process = summarise_results.get_events_to_process(self.yearly_events)
        all_events_temp2 = summarise_results.process_all_events_results(events_to_process)

        all_events_temp2 = summarise_results.join_ewr_parameters(cols_to_add=['Multigauge'],
                        left_table=all_events_temp2,
                        left_on=['gauge','pu','ewr'],
                        selected_columns= ['scenario', 'gauge', 'pu', 'ewr', 'waterYear', 'startDate', 'endDate',
                                'eventDuration', 'eventLength', 
                                'Multigauge'],
                        parameter_sheet_path=self.parameter_sheet)

        all_events_temp2 = summarise_results.filter_duplicate_start_dates(all_events_temp2)


        # Part 1 - Get only the successful events:
        all_successfulEvents = summarise_results.filter_successful_events(all_events_temp2, self.parameter_sheet) 

        # Part 2 - Now we have a dataframe of only successful events, pull down the interevent periods
        # Get start and end date of the timeseries.
        date0 = self.flow_data.index[0]
        date1 = self.flow_data.index[-1]

        start_date = date(date0.year, date0.month, date0.day)
        end_date = date(date1.year, date1.month, date1.day)

        all_successful_interEvents = summarise_results.events_to_interevents(start_date, end_date, all_successfulEvents)

        return all_successful_interEvents



    def get_yearly_ewr_results(self)-> pd.DataFrame:

        if not self.pu_ewr_statistics:
            self.process_scenarios()

        # Setting up the yearly results
        to_process = summarise_results.pu_dfs_to_process(self.pu_ewr_statistics)
        yearly_ewr_results = summarise_results.process_df_results(to_process)
                                
        yearly_ewr_results = summarise_results.join_ewr_parameters(cols_to_add=['Multigauge'],
                                left_table=yearly_ewr_results,
                                left_on=['gauge','pu','ewrCode'],
                                selected_columns= ['Year', 'eventYears', 'numAchieved', 'numEvents', 'numEventsAll',
                                             'eventLength', 'eventLengthAchieved', 'totalEventDays', 'totalEventDaysAchieved',
                                            'maxEventDays', 'maxRollingEvents', 'maxRollingAchievement',
                                            'missingDays', 'totalPossibleDays', 'ewrCode',
                                            'scenario', 'gauge', 'pu', 'Multigauge'],
                                parameter_sheet_path=self.parameter_sheet)

        # Setting up the dictionary of yearly rolling maximum interevent periods:
        events_to_process = summarise_results.get_events_to_process(self.yearly_events)
        all_events_temp = summarise_results.process_all_events_results(events_to_process)

        all_events_temp = summarise_results.join_ewr_parameters(cols_to_add=['Multigauge'],
                        left_table=all_events_temp,
                        left_on=['gauge','pu','ewr'],
                        selected_columns= ['scenario', 'gauge', 'pu', 'ewr', 'waterYear', 'startDate', 'endDate',
                                'eventDuration', 'eventLength', 
                                'Multigauge'],
                        parameter_sheet_path=self.parameter_sheet)
        all_events_temp = summarise_results.filter_duplicate_start_dates(all_events_temp)
        #Filter out the unsuccessful events:
        all_successfulEvents = summarise_results.filter_successful_events(all_events_temp, self.parameter_sheet)

        # Get start and end date of the timeseries.
        date0 = self.flow_data.index[0]
        date1 = self.flow_data.index[-1]
        start_date = date(date0.year, date0.month, date0.day)
        end_date = date(date1.year, date1.month, date1.day)
        df = summarise_results.events_to_interevents(start_date, end_date, all_successfulEvents)
        rolling_max_interevents_dict = summarise_results.get_rolling_max_interEvents(df, start_date, end_date, yearly_ewr_results, self.parameter_sheet)
        # Add the rolling max interevents to the yearly dataframe:
        yearly_ewr_results = summarise_results.add_interevent_to_yearly_results(yearly_ewr_results, rolling_max_interevents_dict)
        
        # Calculate the rolling achievement of the interevent, append this to a new column
        yearly_ewr_results = summarise_results.add_interevent_check_to_yearly_results(yearly_ewr_results, self.parameter_sheet)

        return yearly_ewr_results




    def get_ewr_results(self) -> pd.DataFrame:
        
        if not self.pu_ewr_statistics:
            self.process_scenarios()

        return summarise_results.summarise(self.pu_ewr_statistics , self.yearly_events, parameter_sheet_path=self.parameter_sheet)